/*
* Copyright (c) 2016-2017 Texas Instruments Inc.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without modification,
* are permitted provided that the following conditions are met:
*
* 1. Redistributions of source code must retain the above copyright notice,
*    this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright notice,
*    this list of conditions and the following disclaimer in the documentation
*    and/or other materials provided with the distribution.
* 3. The name of the author may not be used to endorse or promote products
*    derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
* SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
* OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
* IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
* OF SUCH DAMAGE.
*
* This file is part of the lwIP TCP/IP stack.
* 
 * Author: Amit Ashara
*
*/

#ifndef __NETIF_LLDP_H__
#define __NETIF_LLDP_H__

#include "lwip/opt.h"

#if LWIP_LLDP || LWIP_ETHERNET /* don't build if not configured for use in lwipopts.h */

#include "lwip/pbuf.h"
#include "lwip/ip_addr.h"
#include "lwip/netif.h"
#include "lwip/ip.h"

#ifdef __cplusplus
extern "C" {
#endif

/** Header length of LLDP. */
#ifndef ETHLLDP_HEADER_LEN
#define ETHLLDP_HEADER_LEN     14
#endif

/** Mandatory TLVs should be defined. */
#ifndef LLDP_MANDATORY_TLV
#define LLDP_MANDATORY_TLV
#endif

/** Add 2 bytes to the size as the TLV type and length is not a part
 * of the length sub field.  */
#ifndef LLDP_TLV_ADDTOSIZE
#define LLDP_TLV_ADDTOSIZE    2
#endif

/** Maximum size of the pbuf that shall be used for transmitting a
 * lldp packet. */
#ifndef LLDP_PBUF_MAX_SIZE
#define LLDP_PBUF_MAX_SIZE    256
#endif

/** Selection of the multicast destination address. */
#if !defined(LLDP_MULTICAST_ADDR0) && !defined(LLDP_MULTICAST_ADDR1) && !defined(LLDP_MULTICAST_ADDR2)
#define LLDP_MULTICAST_ADDR0   1
#endif

/** Packet structure of the chassis TLV. */
PACK_STRUCT_BEGIN
struct lldp_chassistlv {
  PACK_STRUCT_FIELD(u8_t tlvtype);
  PACK_STRUCT_FIELD(u8_t tlvlen);
  PACK_STRUCT_FIELD(u8_t tlvsubtype);
  PACK_STRUCT_FIELD(u8_t tlvvalue[255]);
} PACK_STRUCT_STRUCT;
PACK_STRUCT_END

/** Packet structure of the Port ID TLV. */
PACK_STRUCT_BEGIN
struct lldp_portidtlv {
  PACK_STRUCT_FIELD(u8_t tlvtype);
  PACK_STRUCT_FIELD(u8_t tlvlen);
  PACK_STRUCT_FIELD(u8_t tlvsubtype);
  PACK_STRUCT_FIELD(u8_t tlvvalue[255]);
} PACK_STRUCT_STRUCT;
PACK_STRUCT_END

/** Packet structure of the Time-to-Live TLV. */
PACK_STRUCT_BEGIN
struct lldp_timetolivetlv {
  PACK_STRUCT_FIELD(u8_t tlvtype);
  PACK_STRUCT_FIELD(u8_t tlvlen);
  PACK_STRUCT_FIELD(u8_t tlvvalue[2]);
} PACK_STRUCT_STRUCT;
PACK_STRUCT_END

/** Packet structure of the End of LLDP TLV. */
PACK_STRUCT_BEGIN
struct lldp_endoflldputlv {
  PACK_STRUCT_FIELD(u8_t tlvtype);
  PACK_STRUCT_FIELD(u8_t tlvlen);
} PACK_STRUCT_STRUCT;
PACK_STRUCT_END

/** Ethernet Type for LLDP. */
#define ETHTYPE_LLDP      0x88CCU

/** Encoding for TLV type field of mandatory TLV's in LLDP. */
#ifdef LLDP_MANDATORY_TLV
#define TLVTYPE_ENDOFLLPDU  0x00U
#define TLVTYPE_CHASSISID   0x01U
#define TLVTYPE_PORTID      0x02U
#define TLVTYPE_TTL         0x03U
#endif

/** Encoding for TLV type field of optional TLV's in LLDP. */
#ifdef LLDP_OPTIONAL_TLV
#define TLVTYPE_PORTDESCR   0x04U
#define TLVTYPE_SYSNAME     0x05U
#define TLVTYPE_SYSDESCR    0x06U
#define TLVTYPE_SYSCAPAB    0x07U
#define TLVTYPE_MGMTADDR    0x08U
#endif

/** Encoding for LLDPMED TLV type. */
#ifdef LLDPMED_TLV
#define TLVTYPE_ORGSPECIFIC 0xEFU
#endif

/** Chassis TLV sub-type definition. */
#define ID_CHASSIS_MACADDR     0x04U
#define ID_CHASSIS_NETADDR     0x05U
#define ID_CHASSIS_LOCALASSIGN 0x07U

/** Port ID TLV sub-type definition. */
#define ID_PORTID_MACADDR     0x03U
#define ID_PORTID_NETADDR     0x04U
#define ID_PORTID_LOCALASSIGN 0x07U

/** Length of the MAC address field. */
#define MAC_ADDRESS_LEN       0x06U

/** Length of IPv4 address field and IANA type. */
#define NETWORK_ADDRESS_IPV4  0x01U
#define IPV4_ADDRESS_LEN      0x04U

/** If the LLDP TTL is not defined then use 120 seconds. */
#ifndef LLDP_TTL_USERVALUE
#define LLDP_TTL_USERVALUE    120
#endif

/** MEMCPY-like macro to copy to/from struct eth_addr's that are no local
 * variables and known to be 16-bit aligned within the protocol header. */
#ifndef ETHLLDP_COPY
#define ETHLLDP_COPY(src, dst, len)  SMEMCPY(src, dst, len)
#endif

/** MEMSET-like macro to initialize a memeory buffer. */
#ifndef ETHLLDP_ZEROINIT
#define ETHLLDP_ZEROINIT(dst, val, len)  SMEMSET(dst, val, len)
#endif

/** Macro to shift the TLV type into a 7-bit octet location bit-1. */
#ifndef LLDPTLVTYPE_COPY
#define LLDPTLVTYPE_COPY(tlvtype) (tlvtype << 1)
#endif

/** Macro to shift the TLV length into a 1-bit octet location bit-0. */
#ifndef LLDPTLVLEN_COPY
#define LLDPTLVLEN_COPY(tlvlen) (u8_t)((tlvlen & 0xFF00) >> 8)
#endif

/** Macro to copy the upper 8 bits of a 16-bit value to a 8-bit buffer. */
#ifndef LLDPOCTET0_COPY
#define LLDPOCTET0_COPY(x) (u8_t)((x & 0xFF00) >> 8)
#endif

/** Macro to copy the lower 8 bits of a 16-bit value to a 8-bit buffer. */
#ifndef LLDPOCTET1_COPY
#define LLDPOCTET1_COPY(x) ((u8_t)x & 0xFF)
#endif

#if LWIP_LLDP /* don't build if not configured for use in lwipopts.h */
err_t lldp_create_chassisidtlv(struct netif *netif, struct pbuf *p, u16_t *len);
err_t lldp_create_portidtlv(struct netif *netif, struct pbuf *p, u16_t *len);
err_t lldp_create_ttltlv(struct netif *netif, struct pbuf *p, u16_t *len);
err_t lldp_create_eoltlv(struct netif *netif, struct pbuf *p, u16_t *len);
err_t lldp_raw(struct netif *netif);
#endif /* LWIP_LLDP */
#endif /* LWIP_LLDP || LWIP_ETHERNET */

#ifdef __cplusplus
}
#endif

#endif /* __NETIF_LLDP_H__ */
