/*
* Copyright (c) 2016-2017 Texas Instruments Inc.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without modification,
* are permitted provided that the following conditions are met:
*
* 1. Redistributions of source code must retain the above copyright notice,
*    this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright notice,
*    this list of conditions and the following disclaimer in the documentation
*    and/or other materials provided with the distribution.
* 3. The name of the author may not be used to endorse or promote products
*    derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
* SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
* OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
* IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
* OF SUCH DAMAGE.
*
* This file is part of the lwIP TCP/IP stack.
* 
 * Author: Amit Ashara
*
*/
 
#include "lwip/opt.h"

#if LWIP_LLDP || LWIP_ETHERNET

#include "lwip/ip_addr.h"
#include "lwip/def.h"
#include "lwip/ip.h"
#include "lwip/stats.h"
#include "lwip/snmp.h"
#include "lwip/dhcp.h"
#include "lwip/autoip.h"
#include "netif/lldp.h"

#include <string.h>

#if LLDP_MULTICAST_ADDR0
const struct eth_addr ethlldpdestmac = {{0x01,0x80,0xC2,0x00,0x00,0x0e}};
#endif
#if LLDP_MULTICAST_ADDR1
const struct eth_addr ethlldpdestmac = {{0x01,0x80,0xC2,0x00,0x00,0x03}};
#endif
#if LLDP_MULTICAST_ADDR2
const struct eth_addr ethlldpdestmac = {{0x01,0x80,0xC2,0x00,0x00,0x00}};
#endif

/**
 * Creates the chassis ID TLV.
 *
 * @param netif the lwIP network interface on which to send the packet
 * @param p the buffer in which the chassis id to be copied
 * @param len is the offset where the tlv needs to be copied and is updated
 * with the offset for the next TLV copy
 *
 * @return
 * - ERR_OK if the chassis ID TLV is created.
 * - ERR_BUF if the final length of the packet exceeds the buffer length.
 * - ERR_MEM if the memory cannot be allocated.
 */
err_t
lldp_create_chassisidtlv(struct netif *netif, struct pbuf *p, u16_t *len)
{
    u16_t u16Len = 0;
    struct lldp_chassistlv *chassis;

    /* Allocate a chassisid tlv with the size of the structure. */
    chassis = (struct lldp_chassistlv *)mem_malloc(sizeof(*chassis));
    if (chassis == NULL) {
        LWIP_DEBUGF(LLDP_DEBUG | LWIP_DBG_TRACE | LWIP_DBG_LEVEL_SERIOUS,
          ("lldp_raw: could not allocate memory for LLDP chassis.\n"));
        LLDP_STATS_INC(lldp.memerr);
        return ERR_MEM;
    }

    /* Initialize the buffer to 0 */
    ETHLLDP_ZEROINIT(chassis, 0, sizeof(*chassis));

    /* If the user has defined the sub type then use the define, else by default
     * the highest priority goes to the MAC address */
#ifdef LLDP_CHASSISID_USERVALUE
#if (LLDP_CHASSISID_USERVALUE == ID_CHASSIS_MACADDR)
    chassis->tlvsubtype = ID_CHASSIS_MACADDR;
    ETHLLDP_COPY(chassis->tlvvalue, &netif->hwaddr, MAC_ADDRESS_LEN);
    u16Len = MAC_ADDRESS_LEN+1;
#elif (LLDP_CHASSISID_USERVALUE == ID_CHASSIS_NETADDR)
    chassis->tlvsubtype = ID_CHASSIS_NETADDR;
    chassis->tlvvalue[0] = NETWORK_ADDRESS_IPV4;
    ETHLLDP_COPY(&chassis->tlvvalue[1], &netif->ip_addr, IPV4_ADDRESS_LEN);
    u16Len = IPV4_ADDRESS_LEN+2;
#elif (LLDP_CHASSISID_USERVALUE == ID_CHASSIS_LOCALASSIGN)
    chassis->tlvsubtype = ID_CHASSIS_LOCALASSIGN;
    ETHLLDP_COPY(chassis->tlvvalue, LLDP_CHASSISID_USERSTRING, sizeof(LLDP_CHASSISID_USERSTRING));
    u16Len = sizeof(LLDP_CHASSISID_USERSTRING)+1;
#else
    chassis->tlvsubtype = ID_CHASSIS_MACADDR;
    ETHLLDP_COPY(chassis->tlvvalue, &netif->hwaddr, MAC_ADDRESS_LEN);
    u16Len = MAC_ADDRESS_LEN+1;
#endif
#else
    chassis->tlvsubtype = ID_CHASSIS_MACADDR;
    ETHLLDP_COPY(chassis->tlvvalue, &netif->hwaddr, MAC_ADDRESS_LEN);
    u16Len = MAC_ADDRESS_LEN+1;
#endif
    /* complete the TLV by putting the type and length field */
    chassis->tlvtype = LLDPTLVTYPE_COPY(TLVTYPE_CHASSISID) | LLDPTLVLEN_COPY(u16Len);
    chassis->tlvlen = LLDPOCTET1_COPY(u16Len);
    u16Len += LLDP_TLV_ADDTOSIZE;

    /* if the TLV addition to the buffer will cause the max size to be exceeded
     * then free the buffer and return a buffer error */
    if(((*len) + u16Len) > LLDP_PBUF_MAX_SIZE)
    {
        mem_free(chassis);
        return ERR_BUF;
    }

    /* Copy the TLV to the pbuf, increment the offset and free the memory
     * allocated to the tlv. */
    ETHLLDP_COPY((u8_t*)p->payload+(*len), chassis, u16Len);
    *len += u16Len;
    mem_free(chassis);

    return ERR_OK;
}

/**
 * Creates the Port ID TLV.
 *
 * @param netif the lwIP network interface on which to send the packet
 * @param p the buffer in which the port id to be copied
 * @param len is the offset where the tlv needs to be copied and is updated
 * with the offset for the next TLV copy
 *
 * @return
 * - ERR_OK if the port ID TLV is created.
 * - ERR_BUF if the final length of the packet exceeds the buffer length.
 * - ERR_MEM if the memory cannot be allocated.
 */
err_t
lldp_create_portidtlv(struct netif *netif, struct pbuf *p, u16_t *len)
{
    u16_t u16Len = 0;
    struct lldp_portidtlv *portid;

    /* Allocate a portid tlv with the size of the structure. */
    portid = (struct lldp_portidtlv *)mem_malloc(sizeof(*portid));
    if (portid == NULL) {
        LWIP_DEBUGF(LLDP_DEBUG | LWIP_DBG_TRACE | LWIP_DBG_LEVEL_SERIOUS,
          ("lldp_raw: could not allocate memory for LLDP portid.\n"));
        LLDP_STATS_INC(lldp.memerr);
        return ERR_MEM;
    }

    /* Initialize the buffer to 0 */
    ETHLLDP_ZEROINIT(portid, 0, sizeof(*portid));

    /* If the user has defined the sub type then use the define, else by default
     * the highest priority goes to the MAC address */
#ifdef LLDP_CHASSISID_USERVALUE
#if (LLDP_PORTID_USERVALUE == ID_PORTID_MACADDR)
    portid->tlvsubtype = ID_PORTID_MACADDR;
    ETHLLDP_COPY(portid->tlvvalue, &netif->hwaddr, MAC_ADDRESS_LEN);
    u16Len = MAC_ADDRESS_LEN+1;
#elif (LLDP_PORTID_USERVALUE == ID_PORTID_NETADDR)
    portid->tlvsubtype = ID_PORTID_NETADDR;
    portid->tlvvalue[0] = NETWORK_ADDRESS_IPV4;
    ETHLLDP_COPY(&portid->tlvvalue[1], &netif->ip_addr, IPV4_ADDRESS_LEN);
    u16Len = IPV4_ADDRESS_LEN+2;
#elif (LLDP_PORTID_USERVALUE == ID_PORTID_LOCALASSIGN)
    portid->tlvsubtype = ID_PORTID_LOCALASSIGN;
    ETHLLDP_COPY(portid->tlvvalue, LLDP_PORTID_USERSTRING, sizeof(LLDP_PORTID_USERSTRING));
    u16Len = sizeof(LLDP_PORTID_USERSTRING)+1;
#else
    portid->tlvsubtype = ID_PORTID_MACADDR;
    ETHLLDP_COPY(portid->tlvvalue, &netif->hwaddr, MAC_ADDRESS_LEN);
    u16Len = MAC_ADDRESS_LEN+1;
#endif
#else
    portid->tlvsubtype = ID_PORTID_MACADDR;
    ETHLLDP_COPY(portid->tlvvalue, &netif->hwaddr, MAC_ADDRESS_LEN);
    u16Len = MAC_ADDRESS_LEN+1;
#endif
    /* complete the TLV by putting the type and length field */
    portid->tlvtype = LLDPTLVTYPE_COPY(TLVTYPE_PORTID) | LLDPTLVLEN_COPY(u16Len);
    portid->tlvlen = LLDPOCTET1_COPY(u16Len);
    u16Len += LLDP_TLV_ADDTOSIZE;

    /* if the TLV addition to the buffer will cause the max size to be exceeded
     * then free the buffer and return a buffer error */
    if(((*len) + u16Len) > LLDP_PBUF_MAX_SIZE)
    {
        mem_free(portid);
        return ERR_BUF;
    }

    /* Copy the TLV to the pbuf, increment the offset and free the memory
     * allocated to the tlv. */
    ETHLLDP_COPY((u8_t*)p->payload+(*len), portid, u16Len);
    *len += u16Len;
    mem_free(portid);

    return ERR_OK;
}

/**
 * Creates the Time-to-Live TLV.
 *
 * @param netif the lwIP network interface on which to send the packet
 * @param p the buffer in which the time-to-live to be copied
 * @param len is the offset where the tlv needs to be copied and is updated
 * with the offset for the next TLV copy
 *
 * @return
 * - ERR_OK if the time-to-live TLV is created.
 * - ERR_BUF if the final length of the packet exceeds the buffer length.
 * - ERR_MEM if the memory cannot be allocated.
 */
err_t
lldp_create_ttltlv(struct netif *netif, struct pbuf *p, u16_t *len)
{
    u16_t u16Len = 0;
    struct lldp_timetolivetlv *ttl;

    /* Allocate a ttl tlv with the size of the structure. */
    ttl = (struct lldp_timetolivetlv *)mem_malloc(sizeof(*ttl));
    if (ttl == NULL) {
        LWIP_DEBUGF(LLDP_DEBUG | LWIP_DBG_TRACE | LWIP_DBG_LEVEL_SERIOUS,
          ("lldp_raw: could not allocate memory for LLDP portid.\n"));
        LLDP_STATS_INC(lldp.memerr);
        return ERR_MEM;
    }

    /* Initialize the buffer to 0 */
    ETHLLDP_ZEROINIT(ttl, 0, sizeof(*ttl));

    /* complete the TLV by putting the value, type and length field */
    ttl->tlvvalue[0] = LLDPOCTET0_COPY(LLDP_TTL_USERVALUE);
    ttl->tlvvalue[1] = LLDPOCTET1_COPY(LLDP_TTL_USERVALUE);
    u16Len = sizeof(ttl->tlvvalue);
    ttl->tlvtype = LLDPTLVTYPE_COPY(TLVTYPE_TTL) | LLDPTLVLEN_COPY(u16Len);
    ttl->tlvlen = LLDPOCTET1_COPY(u16Len);
    u16Len += LLDP_TLV_ADDTOSIZE;

    /* if the TLV addition to the buffer will cause the max size to be exceeded
     * then free the buffer and return a buffer error */
    if(((*len) + u16Len) > LLDP_PBUF_MAX_SIZE)
    {
        mem_free(ttl);
        return ERR_BUF;
    }

    /* Copy the TLV to the pbuf, increment the offset and free the memory
     * allocated to the tlv. */
    ETHLLDP_COPY((u8_t*)p->payload+(*len), ttl, u16Len);
    *len += u16Len;
    mem_free(ttl);

    return ERR_OK;
}

/**
 * Creates the End of LLPDU TLV.
 *
 * @param netif the lwIP network interface on which to send the packet
 * @param p the buffer in which the end of LLPDU to be copied
 * @param len is the offset where the tlv needs to be copied and is updated
 * with the offset for the next TLV copy
 *
 * @return
 * - ERR_OK if the End of LLPDU TLV is created.
 * - ERR_BUF if the final length of the packet exceeds the buffer length.
 * - ERR_MEM if the memory cannot be allocated.
 */
err_t
lldp_create_eoltlv(struct netif *netif, struct pbuf *p, u16_t *len)
{
    u16_t u16Len = 0;
    struct lldp_endoflldputlv *eol;

    /* Allocate a eol with the size of the structure. */
    eol = (struct lldp_endoflldputlv *)mem_malloc(sizeof(*eol));
    if (eol == NULL) {
        LWIP_DEBUGF(LLDP_DEBUG | LWIP_DBG_TRACE | LWIP_DBG_LEVEL_SERIOUS,
          ("lldp_raw: could not allocate memory for LLDP portid.\n"));
        LLDP_STATS_INC(lldp.memerr);
        return ERR_MEM;
    }

    /* Initialize the buffer to 0. This tlv has the value as 0 */
    ETHLLDP_ZEROINIT(eol, 0, sizeof(*eol));
    u16Len = sizeof(*eol);
    u16Len += LLDP_TLV_ADDTOSIZE;

    /* if the TLV addition to the buffer will cause the max size to be exceeded
     * then free the buffer and return a buffer error */
    if(((*len) + u16Len) > LLDP_PBUF_MAX_SIZE)
    {
        mem_free(eol);
        return ERR_BUF;
    }

    /* Copy the TLV to the pbuf, increment the offset and free the memory
     * allocated to the tlv. */
    ETHLLDP_COPY((u8_t*)p->payload+(*len), eol, u16Len);
    *len += u16Len;
    mem_free(eol);

    return ERR_OK;
}

/**
 * Construct and send a raw LLDP packet.
 *
 * @param netif the lwip network interface on which to send the LLDP packet
 *
 * @return ERR_OK if the LLDP packet has been sent
 *         ERR_MEM if the LLDP packet couldn't be allocated
 *         any other err_t on failure
 */
err_t
lldp_raw(struct netif *netif)
{
    err_t result = ERR_OK;
    u16_t u16totallen = 0;
    struct pbuf *pmax;
    struct eth_hdr *ethhdr;

    LWIP_ASSERT("netif != NULL", netif != NULL);

    /* Allocate a pbuf with the maximum size for LLDP packet. */
    pmax = pbuf_alloc(PBUF_RAW, LLDP_PBUF_MAX_SIZE, PBUF_RAM);
    if (pmax == NULL) {
        LWIP_DEBUGF(LLDP_DEBUG | LWIP_DBG_TRACE | LWIP_DBG_LEVEL_SERIOUS,
          ("lldp_raw: could not allocate memory for LLDP pbuf.\n"));
        LLDP_STATS_INC(lldp.memerr);
        return ERR_MEM;
    }

    /* Copy the ethernet source and destination address and ethertype. */
    ethhdr = (struct eth_hdr *)pmax->payload;

    ETHADDR16_COPY(&ethhdr->dest, &ethlldpdestmac);
    ETHADDR16_COPY(&ethhdr->src, netif->hwaddr);
    ethhdr->type = PP_HTONS(ETHTYPE_LLDP);

    /* Update the offset of the next payload data to end of the header.*/
    u16totallen = ETHLLDP_HEADER_LEN;

    /* Call the chassis ID TLV creation */
    result = lldp_create_chassisidtlv(netif, pmax, &u16totallen);
    if(result != ERR_OK)
    {
        LWIP_DEBUGF(LLDP_DEBUG | LWIP_DBG_TRACE | LWIP_DBG_LEVEL_SERIOUS,
          ("lldp_raw: failed chassis tlv creation.\n"));
        LLDP_STATS_INC(lldp.memerr);
        pbuf_free(pmax);
        return ERR_MEM;
    }

    /* Call the port ID TLV creation */
    result = lldp_create_portidtlv(netif, pmax, &u16totallen);
    if(result != ERR_OK)
    {
        LWIP_DEBUGF(LLDP_DEBUG | LWIP_DBG_TRACE | LWIP_DBG_LEVEL_SERIOUS,
          ("lldp_raw: failed port id tlv creation.\n"));
        LLDP_STATS_INC(lldp.memerr);
        pbuf_free(pmax);
        return ERR_MEM;
    }

    /* Call the time-to-live TLV creation */
    result = lldp_create_ttltlv(netif, pmax, &u16totallen);
    if(result != ERR_OK)
    {
        LWIP_DEBUGF(LLDP_DEBUG | LWIP_DBG_TRACE | LWIP_DBG_LEVEL_SERIOUS,
          ("lldp_raw: failed time-to-live tlv creation.\n"));
        LLDP_STATS_INC(lldp.memerr);
        pbuf_free(pmax);
        return ERR_MEM;
    }

    /* Call the end of LLPDU TLV creation */
    result = lldp_create_eoltlv(netif, pmax, &u16totallen);
    if(result != ERR_OK)
    {
        LWIP_DEBUGF(LLDP_DEBUG | LWIP_DBG_TRACE | LWIP_DBG_LEVEL_SERIOUS,
          ("lldp_raw: failed end of lldpu tlv creation.\n"));
        LLDP_STATS_INC(lldp.memerr);
        pbuf_free(pmax);
        return ERR_MEM;
    }

    /* Resize the transmit buffer with the actual length of the LLDP packet. */
    pbuf_realloc(pmax, u16totallen);

    /* send LLDP packet */
    result = netif->linkoutput(netif, pmax);
    LLDP_STATS_INC(lldp.xmit);

    /* free LLDP transmit packet */
    pbuf_free(pmax);
    pmax = NULL;

    return result;
}

#endif /* LWIP_LLDP || LWIP_ETHERNET */
